@@ -12,26 +12,32 @@ from rich import print as rich_print
 from textfsm import TextFSM
 import pexpect as px
 import transitions
+import snoop
 
 """Can't trigger event _go_LOGIN_SUCCESS_UNPRIV from state SEND_LOGIN_PASSWORD!"""
 
 class UnexpectedPrompt(Exception):
     """Exception for an Unexpected Prompt"""
-    def __init__(self):
-        super(UnexpectedPrompt, self)
+    def __init__(self, error=""):
+        super(UnexpectedPrompt, self).__init__(error)
 
 class PromptDetectionError(Exception):
     """Exception for an Unexpected Prompt"""
-    def __init__(self):
-        super(PromptDetectionError, self)
+    def __init__(self, error=""):
+        super(PromptDetectionError, self).__init__(error)
+
+# DO NOT escape '$' here
+EXPECTED_LAST_PROMPT_CHARS = (':', '>', '#', '$')
 
 class TeeStdoutFile(object):
     """Simple class to send stdout to screen and log_file simultaneously"""
-    def __init__(self, log_file="", filemode="w", log_screen=False):
+    def __init__(self, log_file="", filemode="w", log_screen=False,
+        encoding='utf-8'):
         self.log_file = os.path.expanduser(log_file)
         self.filemode = filemode
         self.log_screen = log_screen
         self.stdout = sys.stdout
+        self.encoding = encoding
 
         try:
             assert os.path.isfile(self.log_file) is False
@@ -40,7 +46,7 @@ class TeeStdoutFile(object):
                 self.log_file))
             sys.exit(1)
 
-        self.fh = open(self.log_file, self.filemode)
+        self.fh = open(self.log_file, self.filemode, encoding=self.encoding)
 
     def __del__(self):
         sys.stdout = self.stdout
@@ -51,8 +57,13 @@ class TeeStdoutFile(object):
             pass
 
     def write(self, line):
-        self.fh.write(line)
-        self.stdout.write(line)
+        for ii in line:
+            if isinstance(ii, bytes):
+                self.fh.write(ii.decode(self.encoding))
+                self.stdout.write(ii.decode(self.encoding))
+            elif isinstance(ii, str):
+                self.fh.write(ii)
+                self.stdout.write(ii)
 
     def flush(self):
         self.fh.flush()
@@ -74,7 +85,7 @@ class Account(object):
 class Shell(transitions.Machine):
     def __init__(self, host='', credentials=(), 
         ssh_keepalive=60, protocols=({'proto': 'ssh', 'port': 22},
-        {'proto': 'telnet', 'port': 23}), auto_priv_mode=True,
+        {'proto': 'telnet', 'port': 23}), auto_priv_mode=None,
         log_screen=False, log_file='', debug=False, command_timeout=30,
         login_timeout=10, relogin_delay=120, encoding='utf-8',
         login_attempts=3):
@@ -84,13 +95,16 @@ class Shell(transitions.Machine):
             'SEND_LOGIN_PASSWORD', 'CONNECT', 
             'LOGIN_SUCCESS_UNPRIV', 'LOGIN_SUCCESS_PRIV',
             'LOGIN_TIMEOUT', 'SEND_PRIV_PASSWORD', 
-            'INTERACT', 'CLOSE_SESSION')
+            'LOGIN_COMPLETE', 'INTERACT', 'CLOSE_SESSION')
         super(Shell, self).__init__(states=STATES, initial='INIT_SESSION')
 
         self.host = host
         self.credentials = credentials
         self.protocols = protocols
         self.auto_priv_mode = auto_priv_mode
+        if auto_priv_mode is not None:
+            rich_print("[bold red]WARNING: auto_priv_mode will be deprecated.  Please stop using this option.[/bold red]")
+            time.sleep(2.5)
         self.log_screen = log_screen
         self.log_file = os.path.expanduser(log_file)
         self.debug = debug
@@ -111,15 +125,17 @@ class Shell(transitions.Machine):
         self.prompt_str = ""    # This gets set in self.sync_prompt()
         # Detect a typical linux CLI prompt...
         #self.linux_prompt_capture = '([^\r\n{0}]+)\s*$'.format(re.escape('$'))
-        self.linux_prompt_capture = '[\r\n]*[^\r\n{0}]+\s*$'.format(re.escape('$'))
+        self.linux_prompt_capture = '[\r\n]*[^\r\n{0}]+\s*{0}'.format(re.escape('$'))
         # Build the template before detecting prompt
-        self.base_prompt_regex = self.build_prompt_regex()
+        self.base_prompt_regex = self.build_base_prompt_regex()
 
         # Define regex capture groups for the prompts above...
         # NOTE there are no prompts in these strings..
         self.base_prompt_regex_capture = [':', ':', 
             '(\S[^\n\r>]+?)\s*$', self.linux_prompt_capture, 
             '(\S[^\n\r{0}]+?)\s*$'.format(re.escape('#'))]
+            #'(\S[^\n\r>]+?)\s*>', self.linux_prompt_capture, 
+            #'(\S[^\n\r{0}]+?)\s*{0}'.format(re.escape('#'))]
 
         self.matching_prompt_regex = ""
         self.matching_prompt_regex_index = -1
@@ -142,6 +158,10 @@ class Shell(transitions.Machine):
             source='SEND_LOGIN_PASSWORD', dest='SELECT_LOGIN_CREDENTIALS',
                 after='after_SELECT_LOGIN_CREDENTIALS_cb')
 
+        self.add_transition(trigger='_go_SELECT_LOGIN_CREDENTIALS', 
+            source='LOGIN_COMPLETE', dest='SELECT_LOGIN_CREDENTIALS',
+                after='after_SELECT_LOGIN_CREDENTIALS_cb')
+
         #######################################################################
         ## Transitions to CONNECT state
         #######################################################################
@@ -154,15 +174,15 @@ class Shell(transitions.Machine):
             after='after_CONNECT_cb')
 
         #######################################################################
-        ## Transitions to SEND_LOGIN_PASSWORD state
+        ## Transitions to SEND_LOGIN_USERNAME state
         #######################################################################
         self.add_transition(trigger='_go_SEND_LOGIN_USERNAME', 
             source='CONNECT', dest='SEND_LOGIN_USERNAME',
             after='after_SEND_LOGIN_USERNAME_cb')
 
-        # In case we got to interact prematurely...
+        # In case we got to LOGIN_COMPLETE prematurely...
         self.add_transition(trigger='_go_SEND_LOGIN_USERNAME', 
-            source='INTERACT', dest='SEND_LOGIN_USERNAME',
+            source='LOGIN_COMPLETE', dest='SEND_LOGIN_USERNAME',
             after='after_SEND_LOGIN_USERNAME_cb')
 
         #######################################################################
@@ -172,9 +192,13 @@ class Shell(transitions.Machine):
             source='CONNECT', dest='SEND_LOGIN_PASSWORD',
             after='after_SEND_LOGIN_PASSWORD_cb')
 
-        # In case we got to interact prematurely...
         self.add_transition(trigger='_go_SEND_LOGIN_PASSWORD', 
-            source='INTERACT', dest='SEND_LOGIN_PASSWORD',
+            source='SEND_LOGIN_USERNAME', dest='SEND_LOGIN_PASSWORD',
+            after='after_SEND_LOGIN_PASSWORD_cb')
+
+        # In case we got to LOGIN_COMPLETE prematurely...
+        self.add_transition(trigger='_go_SEND_LOGIN_PASSWORD', 
+            source='LOGIN_COMPLETE', dest='SEND_LOGIN_PASSWORD',
             after='after_SEND_LOGIN_PASSWORD_cb')
 
         #######################################################################
@@ -210,19 +234,26 @@ class Shell(transitions.Machine):
             source='SEND_LOGIN_PASSWORD', dest='LOGIN_SUCCESS_PRIV',
             after='after_LOGIN_SUCCESS_PRIV_cb')
 
+        #######################################################################
+        ## Transitions to LOGIN_COMPLETE state
+        #######################################################################
+        self.add_transition(trigger='_go_LOGIN_COMPLETE', 
+            source='LOGIN_SUCCESS_UNPRIV', dest='LOGIN_COMPLETE',
+            after='after_LOGIN_COMPLETE_cb')
+
+        self.add_transition(trigger='_go_LOGIN_COMPLETE', 
+            source='LOGIN_SUCCESS_PRIV', dest='LOGIN_COMPLETE',
+            after='after_LOGIN_COMPLETE_cb')
+
+        self.add_transition(trigger='_go_LOGIN_COMPLETE', 
+            source='LOGIN_COMPLETE', dest='LOGIN_COMPLETE',
+            after='after_LOGIN_COMPLETE_cb')
+
         #######################################################################
         ## Transitions to INTERACT state
         #######################################################################
         self.add_transition(trigger='_go_INTERACT', 
-            source='LOGIN_SUCCESS_UNPRIV', dest='INTERACT',
-            after='after_INTERACT_cb')
-
-        self.add_transition(trigger='_go_INTERACT', 
-            source='LOGIN_SUCCESS_PRIV', dest='INTERACT',
-            after='after_INTERACT_cb')
-
-        self.add_transition(trigger='_go_INTERACT', 
-            source='INTERACT', dest='INTERACT',
+            source='LOGIN_COMPLETE', dest='INTERACT',
             after='after_INTERACT_cb')
 
         #######################################################################
@@ -244,6 +275,10 @@ class Shell(transitions.Machine):
         it returns None.
         """
         assert cmd is not None
+        assert self.child.isalive()  # Don't issue commands against a dead conn
+
+        if self.debug:
+            rich_print("[bold blue]execute() called[/bold blue]")
 
         if command_timeout==0.0:
             command_timeout = self.command_timeout
@@ -275,9 +310,10 @@ class Shell(transitions.Machine):
             self.child.send(cmd)
 
         # Extend the list of cli_prompts if `prompts` was specified
-        cli_prompts = self.base_prompt_regex
-        # Ensure we match sudo first... this regex is tricky to get right...
-        cli_prompts.insert(0, '[Pp]assword[^\r\n]{0,20}?:')
+        # FIXME
+        #cli_prompts = self.base_prompt_regex
+        cli_prompts = self.build_base_prompt_regex()
+
         if prompts!=():
             assert isinstance(prompts, tuple)
             cli_prompts.extend(prompts)  # Add command-specific prompts here...
@@ -286,10 +322,10 @@ class Shell(transitions.Machine):
         index = self.cexpect(cli_prompts, timeout=command_timeout)
 
         # Handle sudo password prompt...
-        if index==0:  # a password prompt
+        if index==1:  # a password prompt
             if self.debug:
                 rich_print("    [bold blue]Responding to password prompt:[/bold blue] [bold yellow]'{}'[/bold yellow]".format(
-                    self.child.after.splitlines()[-1]))
+                    self.matching_prompt))
             assert self.password!="", "Hit a password prompt without a password"
             self.csendline(self.password)
             index = self.cexpect(cli_prompts, timeout=command_timeout)
@@ -336,23 +372,32 @@ class Shell(transitions.Machine):
     def csendline(self, text):
         assert self.child.isalive()
         if self.debug:
-            rich_print("    [bold blue]csendline('{}')[/bold blue]".format(
+            rich_print("    [bold blue]csendline('{}') called[/bold blue]".format(
                 text))
             rich_print("    [bold blue]---------[/bold blue]")
+        # WARNING: use self.child.sendline(); do not use self.csendline() here
         self.child.sendline(text)
 
     def cexpect(self, pattern_list, timeout=-1):
         assert self.child.isalive()
         if self.debug:
-            rich_print("    [bold blue]cexpect([/bold blue][bold green]pattern_list, timeout={}[/bold green][bold blue])[/bold blue]".format(timeout))
+
+            rich_print("    [bold blue]cexpect([/bold blue][bold green]pattern_list, timeout={}[/bold green][bold blue]) called[/bold blue]".format(timeout))
+            # Expand all pattern_list terms...
+            rich_print("        [bold blue]pattern_list = {}[/bold blue]".format("[["))
+            for idx, term in enumerate(pattern_list):
+                rich_print("            [bold blue]{} {},[/bold blue]".format(
+                    idx, repr(term)))
+            rich_print("        [bold blue]{}[/bold blue]".format(str("]]")))
+
         now = time.time()
         try:
             match_index = self.child.expect(pattern_list, timeout=timeout)
             if self.debug:
                 delta_secs = round(time.time()-now, 4)
-                rich_print("\n      [bold blue]^^ cexpect() matched pattern_list index={}[/bold blue]".format(match_index))
-                rich_print("\n      [bold blue]^^ cexpect() matched regex={}[/bold blue]".format(repr(pattern_list[match_index])))
-                rich_print("      [bold blue]^^ cexpect() matching_prompt='{}'[/bold blue]".format(self.matching_prompt))
+                rich_print("\n      [bold blue]^^ cexpect() matched regex at match_index {}={}[/bold blue]".format(match_index, repr(pattern_list[match_index])))
+                # FIXME why can't I use quotes around self.matching_prompt in rich_print()
+                rich_print("      [bold blue]^^ cexpect() matching_prompt={}[/bold blue]".format(repr(self.matching_prompt)))
                 rich_print("      [bold blue]^^ cexpect() match time: {} seconds[/bold blue]".format(delta_secs))
 
             self.matching_prompt_regex_index = match_index
@@ -360,24 +405,39 @@ class Shell(transitions.Machine):
 
         except px.exceptions.EOF:
             if self.debug:
-                rich_print("      [bold red]EOF exception[/bold red]")
+                rich_print("      [bold red]cexpect() EOF exception while waiting for pattern_list[/bold red]")
             match_index = -1
 
         except px.exceptions.TIMEOUT:
             if self.debug:
-                rich_print("      [bold red]TIMEOUT exception[/bold red]")
+                rich_print("      [bold red]cexpect() TIMEOUT exception while waiting for pattern_list[/bold red]")
             match_index = -1
 
         return match_index
 
+    def interact(self):
+        self._go_INTERACT()
+
     @property
     def matching_prompt(self):
-        """Derive the matching prompt string (what matched the regex)"""
-        after = self.child.after
-        prompt_str = after.splitlines()[-1]
+        """Get the matching prompt (what matched the regex, no regex chars)"""
+        assert self.child.isalive()
+
         if self.debug:
-            rich_print("[bold blue]matching_prompt(): '{}'[/bold blue]".format(prompt_str))
-        return prompt_str
+            rich_print("[bold blue]matching_prompt property[/bold blue]")
+        after = self.child.after
+        candidate_prompt = after.splitlines()[-1].strip()   # Must use strip() here...
+
+        match_bool = False
+        for expected_last_char in EXPECTED_LAST_PROMPT_CHARS:
+            if (candidate_prompt[-1]==expected_last_char):
+                match_bool = True
+
+        assert match_bool is True, "Can't find last prompt character in candidate_prompt: {}".format(candidate_prompt)
+        if self.debug:
+            rich_print("    [bold blue]candidate_prompt should have: {}[/bold blue]".format(EXPECTED_LAST_PROMPT_CHARS))
+            rich_print("    [bold blue]candidate_prompt: '{}'[/bold blue]".format(candidate_prompt))
+        return candidate_prompt
 
     @property
     def response(self):
@@ -386,9 +446,6 @@ class Shell(transitions.Machine):
     def exit(self):
         self.child.close()
 
-    def close(self):
-        self.exit()
-
     def quit(self):
         self.exit()
 
@@ -404,115 +461,195 @@ class Shell(transitions.Machine):
                     repr(cred)))
             yield cred
 
-    def sync_prompt(self):
+    def sync_prompt(self, require_detect_prompt=True):
         """Catch up with any queued prompts, we know to exit if we get a px.exceptions.TIMEOUT error"""
         if self.debug:
-            rich_print("[bold blue]Entering sync_prompt()[/bold blue]")
+            rich_print("[bold blue]sync_prompt(require_detect_prompt={}) called[/bold blue]".format(require_detect_prompt))
 
-        # self.detect_prompt() **must** come before self.sync_prompt()
-        try:
-            assert self.prompt_str!=""
-        except AssertionError:
+        # self.detect_prompt() *should* come before self.sync_prompt()
+        if (self.prompt_str=="") and (require_detect_prompt is True):
             raise PromptDetectionError("detect_prompt() must run before sync_prompt()")
 
-        self.csendline('')
+        # WARNING: use self.child.sendline(); do not use self.csendline() here
+        self.child.sendline('')
 
         finished = False
         while not finished:
             # Use a very short timeout here...
-            index = self.cexpect(self.base_prompt_regex, timeout=1)
+            # WARNING self.child.expect() is required; do not use self.cexpect()
+            try:
+                index = self.child.expect(self.base_prompt_regex, timeout=1)
+                if self.debug:
+                    rich_print("    [bold blue]sync_prompt() index={}[/bold blue]".format(index))
+            except px.exceptions.TIMEOUT:
+                # We got an EOF or TIMEOUT error...
+                self.login_attempts = 0
+                finished = True
+            except px.exceptions.EOF:
+                # We got an EOF or TIMEOUT error...
+                self.login_attempts = 0
+                finished = True
 
             if index==-1:
                 # We got an EOF or TIMEOUT error...
                 self.login_attempts = 0
                 finished = True
 
-            elif (index==0 or index==1):
-                raise Exception("Unexpected prompt in sync_prompt()")
+            if (index==0 or index==1):
+                raise UnexpectedPrompt("Unexpected prompt in sync_prompt(): {}".format(self.matching_prompt))
 
             elif index==2:
                 # We should only get to this prompt if auto_priv_mode is 
                 #     False
-                assert self.auto_priv_mode is False
                 self.login_attempts = 0
             elif index==3:
                 # We should only get to this prompt if auto_priv_mode is 
                 #     False
-                assert self.auto_priv_mode is False
                 self.login_attempts = 0
             elif index==4:
                 # We don't need to attempt any more logins if we have 
                 #     a priv prompt
                 self.login_attempts = 0
 
+
+    def close(self):
+        self.child.close()
+        return (self.child.exitstatus, self.child.signalstatus)
+
     def detect_prompt(self):
-        """detect_prompt() checks for premature entry into INTERACT and also looks for a prompt string"""
+        """detect_prompt() checks for premature entry into LOGIN_COMPLETE and also looks for a prompt string"""
         # Detect the prompt as best-possible...
         if self.debug:
-            rich_print("[bold blue]Entering detect_prompt()[/bold blue]")
+            rich_print("[bold blue]detect_prompt() called[/bold blue]")
 
-        abbv_prompt_list = ['sername:', 'assword:', '>', re.escape('$'), 
+        abbv_prompt_list = [r'sername:', r'[Pp]assword[^\r\n]{0,20}?:', r'>', re.escape('$'), 
             re.escape('#')]
 
+        #self.csendline('')  # FIXME I might need to delete this...
+
         # Double check that this isn't a pre-login banner...
+        ii = 0
         finished = False
         while not finished:
             # Use a very short timeout here...
-            index = self.cexpect(abbv_prompt_list, timeout=1)
-            if index==-1:
+            # WARNING self.child.expect() is required; do not use self.cexpect()
+            try:
+                ii += 1  # Keep track of how many times we loop through input
+                index = self.child.expect(abbv_prompt_list, timeout=1)
+            except px.exceptions.TIMEOUT:
+                assert self.child.isalive()
                 if self.debug:
-                    rich_print("[bold blue]    detect_prompt() TIMEOUT condition 1[/bold blue]")
+                    rich_print("[bold blue]    detect_prompt() pre-login finished because it found TIMEOUT condition while looking for abbv_prompt_list[/bold blue]")
+                index = -1
                 finished = True
+            except px.exceptions.EOF:
+                if self.debug:
+                    rich_print("[bold blue]    detect_prompt() pre-login finished because it found EOF condition while looking for abbv_prompt_list[/bold blue]")
+                index = -1
+                finished = True
+
+            if index==-1:
+                assert self.child.isalive()
+                if self.debug:
+                    rich_print("[bold blue]    detect_prompt() index={}[/bold blue]".format(index))
+
             elif index==0:
+                assert self.child.isalive()
                 # We probably went to detect_prompt() based on some banner
                 # This should be a username prompt...
                 if self.debug:
                     rich_print("[bold blue]    detect_prompt() found a premature entry into detect_prompt().  Redirecting to state SEND_LOGIN_USERNAME[/bold blue]")
                 self._go_SEND_LOGIN_USERNAME()
+
             elif index==1:
+                assert self.child.isalive()
                 # We probably went to detect_prompt() based on some banner
                 # This should be a password prompt...
                 if self.debug:
                     rich_print("[bold blue]    detect_prompt() found a premature entry into detect_prompt().  Redirecting to state SEND_LOGIN_PASSWORD[/bold blue]")
+
                 self._go_SEND_LOGIN_PASSWORD()
 
-            ## FIXME I might have to delete this case... probably breaks junos
-            ## with a login banner (using hash characters)
+            elif index>1:
+                # We got a non-user, non-password prompt... exit this loop!
+                # Start prompt detection heuristics...
+                if self.debug:
+                    finished = True
 
-            ## GUACAMOLE
-            #elif index>1:
-            #    finished = True
+            if self.debug:
+                rich_print("    [bold blue]detect_prompt() loop={}".format(ii))
 
         if self.debug:
-            rich_print("\n    [bold blue]detect_prompt() using prompt detection heuristics[/bold blue]")
-        self.csendline('')
+            rich_print("\n    [bold blue]detect_prompt() finished the while loop after {} iterations. Detected prompt index={}[/bold blue]".format(ii, index))
+            rich_print("    [bold blue]detect_prompt() using prompt detection heuristics[/bold blue]")
+        # WARNING: use self.child.sendline(); do not use self.csendline() here
+        self.child.sendline('')
+
+        ### Start building prompt_str and base_prompt_regex
         # Use a very short timeout here
-        index = self.cexpect(abbv_prompt_list, timeout=1)
-        self.prompt_str = self.child.before.strip().splitlines()[-1]
-        if self.debug:
-            rich_print("\n    [bold blue]detect_prompt() found prompt_str:[/bold blue] '{}'".format(self.prompt_str))
+        # WARNING self.child.expect() is required; do not use self.cexpect()
+        try:
+            # replaced self.base_prompt_regex with abbv_prompt_list
+            index = self.child.expect(abbv_prompt_list, timeout=1)
+        except px.exceptions.TIMEOUT:
+            assert self.child.isalive()
+            if self.debug:
+                rich_print("[bold blue]    detect_prompt() pre-login finished.  Hit TIMEOUT condition[/bold blue]")
+        except px.exceptions.EOF:
+            assert self.child.isalive()
+            if self.debug:
+                rich_print("[bold blue]    detect_prompt() pre-login finished.  Hit EOF condition[/bold blue]")
 
-        self.build_prompt_regex()  # Adjust the prompt regex after detection
+        ## Example of prompt detection on route-views.oregon-ix.org...
+        hostname = self.child.before.strip()  # detect hostname    = route-views
+        prompt_char = self.child.after.strip()# detect prompt_char = >
+        hostname_output_list = hostname.splitlines()
+        if len(hostname_output_list)==0:
+            prompt_str = ""   # This is bad... we don't want to hit this...
+        elif len(hostname_output_list)==1:
+            prompt_str = hostname_output_list[0]
+        elif len(hostname_output_list)>1:
+            prompt_str = hostname_output_list[-1] # Get the last entry in list
+        else:
+            raise PromptDetectionError()
+        self.prompt_str = prompt_str
+
+        if self.debug:
+            rich_print("\n        [bold blue]detect_prompt() found prompt_str:[/bold blue] '{}'".format(self.prompt_str))
+
+        self.build_base_prompt_regex()  # Adjust the prompt regex after detection
 
         return self.prompt_str
 
-    def build_prompt_regex(self):
+    def build_base_prompt_regex(self):
         """Assign self.base_prompt_regex with the latest prompt info"""
         if self.debug:
-            rich_print("[bold blue]build_prompt_regex() called[/bold blue]")
+            rich_print("[bold blue]build_base_prompt_regex() called[/bold blue]")
+            rich_print("    [bold blue]self.prompt_str should not have the ending character (like ':', '>', '#', '$')[/bold blue]")
             rich_print("    [bold blue]self.prompt_str='{}'[/bold blue]".format(
                 self.prompt_str))
 
-        self.linux_prompt = '[\n\r]+{0}[^{1}]*?{1}\s*'.format(re.escape(
+        if self.debug:
+            rich_print("    [bold blue]while not end_loop_bool[/bold blue]")
+
+        self.linux_prompt = r'[\n\r]+{0}[^{1}]*?{1}\s*'.format(re.escape(
             self.prompt_str), re.escape('$'))
 
-        self.base_prompt_regex = ['assword:', 'sername:', 
+        self.base_prompt_regex = [r'sername:', r'[Pp]assword[^\r\n]{0,20}?:',
             #r'[\n\r]+{}[^\n\r>]*?>\s*'.format(self.prompt_str), 
-            '[\n\r]+{0}.*?>\s*'.format(re.escape(self.prompt_str)),
+            r'[\n\r]+{0}.*?>\s*'.format(re.escape(self.prompt_str)),
             #linux_prompt, r'[\n\r]+{0}[^\n\r#]*?{1}\s*'.format(
-            self.linux_prompt, '[\n\r]+{0}.*?{1}\s*'.format(
+            self.linux_prompt, r'[\n\r]+{0}.*?{1}\s*'.format(
             re.escape(self.prompt_str), re.escape('#'))]
 
+        if self.debug:
+            # Expand all base_prompt_regex terms...
+            rich_print("        [bold blue]self.base_prompt_regex = {}[/bold blue]".format("[["))
+            for idx, term in enumerate(self.base_prompt_regex):
+                rich_print("            [bold blue]{} {},[/bold blue]".format(
+                    idx, repr(term)))
+            rich_print("        [bold blue]{}[/bold blue]".format(str("]]")))
+
         return self.base_prompt_regex
 
     def after_SELECT_PROTOCOL_cb(self):
@@ -567,7 +704,7 @@ class Shell(transitions.Machine):
             self.ssh_key = cred.ssh_key
 
             if self.child is not None:
-                self.child.close()      # Make way for a fresh child instance
+                self.close()      # Make way for a fresh child instance
 
             self._go_CONNECT()
 
@@ -581,7 +718,7 @@ class Shell(transitions.Machine):
                 self.state))
 
         if self.child is not None:
-            self.child.close()
+            self.close()
 
         # Implement ssh or telnet command...
         if self.proto_dict['proto']=='ssh' and self.ssh_key!="":
@@ -607,7 +744,6 @@ class Shell(transitions.Machine):
         except px.exceptions.EOF:
             time.sleep(70)
 
-
         # log to screen if requested
         if self.log_screen and self.log_file=="":
              self.child.logfile = sys.stdout
@@ -621,19 +757,21 @@ class Shell(transitions.Machine):
             self.child.logfile = TeeStdoutFile(log_file=self.log_file,
             log_screen=self.log_screen)
 
-        index = self.cexpect(self.base_prompt_regex, 
-            timeout=self.login_timeout)
+        if self.debug:
+            rich_print("    [bold blue]Call cexpect() from after_CONNECT_cb()[/bold blue]")
+        index = self.cexpect(self.base_prompt_regex, timeout=self.login_timeout)
 
         if self.debug:
-            rich_print("   [bold blue]after_CONNECT_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(index))
+            rich_print("       [bold blue]self.cexpect() matched prompt index:[/bold blue] {}".format(index))
+            rich_print("       [bold blue]self.cexpect() matching prompt:[/bold blue] {}".format(self.base_prompt_regex[index]))
 
         if index==-1:
             time.sleep(70)
             self._go_CONNECT()
         elif index==0:
-            self._go_SEND_LOGIN_PASSWORD()
-        elif index==1:
             self._go_SEND_LOGIN_USERNAME()
+        elif index==1:
+            self._go_SEND_LOGIN_PASSWORD()
         elif index==2:
             self._go_LOGIN_SUCCESS_UNPRIV()
         elif index==3:
@@ -655,15 +793,14 @@ class Shell(transitions.Machine):
 
             if self.debug:
                 rich_print("\n   [bold blue]after_SEND_LOGIN_PASSWORD_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(index))
+                rich_print("\n   [bold blue]after_SEND_LOGIN_PASSWORD_cb() - self.child.expect() matching_ prompt:[/bold blue] {}".format(repr(self.base_prompt_regex[index])))
 
             if index==0:  # This was the wrong password
-                self.child.close()
+                self.close()
+                self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
+            elif index==1:
+                self.close()
                 self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
-            elif index==1:  # Strange event... should not see username entry here...
-                raise UnexpectedPrompt(
-                    # FIXME
-                    "UnexpectedPrompt: '{}' - after_SEND_LOGIN_PASSWORD_cb".format(""
-                    ))
             elif index==2:
                 self._go_LOGIN_SUCCESS_UNPRIV()  # We got an unpriv prompt here
             elif index==3:
@@ -674,14 +811,16 @@ class Shell(transitions.Machine):
         except px.exceptions.EOF:
             if self.debug:
                 rich_print("   [bold red]after_SEND_LOGIN_PASSWORD_cb() - pexpect.exceptions.EOF error[/bold red]")
-            self.child.close()
-            self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
+            self.close()
+            #self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
+            self._go_CONNECT()  # Restart login with different creds
 
         except px.exceptions.TIMEOUT:
             if self.debug:
                 rich_print("   [bold red]after_SEND_LOGIN_PASSWORD_cb() - pexpect error: TIMEOUT[/bold red]")
-            self.child.close()
-            self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
+            self.close()
+            #self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
+            self._go_CONNECT()  # Restart login with different creds
 
     def after_SEND_LOGIN_USERNAME_cb(self):
 
@@ -696,11 +835,12 @@ class Shell(transitions.Machine):
 
         if self.debug:
             rich_print("   [bold blue]after_SEND_LOGIN_USERNAME_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(index))
+            rich_print("   [bold blue]after_SEND_LOGIN_USERNAME_cb() - self.child.expect() matching_prompt:[/bold blue] {}".format(self.base_prompt_regex[index]))
 
         if index==0:
-            self._go_SEND_LOGIN_PASSWORD()
-        elif index==1:
             self._go_SELECT_LOGIN_CREDENTIALS()
+        elif index==1:
+            self._go_SEND_LOGIN_PASSWORD()
         elif index==2:
             self._go_LOGIN_SUCCESS_UNPRIV()  # We got an unpriv prompt here
         elif index==3:
@@ -714,27 +854,27 @@ class Shell(transitions.Machine):
             rich_print("[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                 self.state))
 
-        if self.auto_priv_mode is True:
-            self.csendline('enable')
+        self.csendline('')
 
-            index = self.child.expect(self.base_prompt_regex,
-                timeout=self.command_timeout)
+        index = self.cexpect(self.base_prompt_regex,
+            timeout=self.command_timeout)
 
-            if self.debug:
-                rich_print("   [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(index))
+        if self.debug:
+            rich_print("   [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(index))
+            rich_print("   [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() - self.child.expect() matching_prompt:[/bold blue] {}".format(self.base_prompt_regex[index]))
 
-            if index==0:
-                self._go_SEND_PRIV_PASSWORD()
-            elif index==1:
-                raise Exception("Unexpected prompt: 'name:'")
-            elif index==2:
-                raise Exception("Unexpected prompt: '>'")
-            elif index==3:
-                raise Exception("Unexpected prompt: '$'")
-            elif index==4:
-                self._go_LOGIN_SUCCESS_PRIV()    # We got a priv prompt here
+        if index==0:
+            raise Exception("Unexpected prompt: 'name:'")
+        elif index==1:
+            self._go_SEND_PRIV_PASSWORD()
+        elif index==2:
+            pass # Got a '>' prompt
+        elif index==3:
+            pass # Got a '#' prompt
+        elif index==4:
+            self._go_LOGIN_SUCCESS_PRIV()    # We got a priv prompt here
 
-        self._go_INTERACT()
+        self._go_LOGIN_COMPLETE()
 
     def after_SEND_PRIV_PASSWORD_cb(self):
 
@@ -749,15 +889,16 @@ class Shell(transitions.Machine):
 
         if self.debug:
             rich_print("   [bold blue]after_SEND_PRIV_PASSWORD_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(index))
+            rich_print("   [bold blue]after_SEND_PRIV_PASSWORD_cb() - self.child.expect() matching_prompt:[/bold blue] {}".format(self.base_prompt_regex[index]))
 
         if index==0:
-            self._go_SEND_PRIV_PASSWORD()
+            raise UnexpectedPrompt("Unexpected prompt: 'name:'")
         elif index==1:
-            raise Exception("Unexpected prompt: 'name:'")
+            self._go_SEND_PRIV_PASSWORD()
         elif index==2:
-            raise Exception("Unexpected prompt: '>'")
+            raise UnexpectedPrompt("Unexpected prompt: '>'")
         elif index==3:
-            raise Exception("Unexpected prompt: '$'")
+            raise UnexpectedPrompt("Unexpected prompt: '$'")
         elif index==4:
             self._go_LOGIN_SUCCESS_PRIV()    # We got a priv prompt here
 
@@ -767,18 +908,33 @@ class Shell(transitions.Machine):
             rich_print("[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                 self.state))
 
-        self._go_INTERACT()
+        self._go_LOGIN_COMPLETE()
 
-    def after_INTERACT_cb(self):
-        """Catch up on any queued command prompts and prepare for user to interact"""
+    def after_LOGIN_COMPLETE_cb(self):
+        """Clean up on any queued command prompts and accept user commands"""
 
         if self.debug:
             rich_print("[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                 self.state))
 
-        self.detect_prompt()  # detect_prompt() *must* come before sync_prompt()
+        #if self.debug:
+        #    rich_print("    [bold blue]Call detect_prompt() from LOGIN_COMPLETE[/bold blue]")
+        if self.debug:
+            rich_print("    [bold blue]Calling sync_prompt() from after_LOGIN_COMPLETE_cb()[/bold blue]")
+        self.detect_prompt() # detect_prompt() *should* come before sync_prompt()
         self.sync_prompt()
 
+    def after_INTERACT_cb(self):
+        """Allow unscripted interaction with the system"""
+        assert self.child.isalive()
+
+        if self.debug:
+            rich_print("[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
+                self.state))
+        rich_print("[bold magenta]Enter INTERACT mode; use Cntl-] to escape[/bold magenta]")
+        self.child.interact(escape_character=chr(4),
+            input_filter=None, output_filter=interact_filter)
+
 
 if __name__=='__main__':
     sess = Shell(
